#
# pushare - P2P file sharing system
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: $
#

package Pushare::Cache;
use IPC::Open2;
use Jcode;
use Pushare::Config;

$cachedir = $Pushare::Config::cachedir;
$updir	  = $Pushare::Config::updir;
$downdir  = $Pushare::Config::downdir;
$namelist = $Pushare::Config::namelist;

sub list(;%) {
	my %opt  = @_;
	my %name = ();
	local $_;
	Pushare::Util::lock() if ($opt{lock});
	if (open IN, $namelist) {
		while (<IN>) {
			chomp;
			local @_ = split /<>/, $_, 2;
			$name{$_[0]} = $_[1];
		}
		close IN;
	}
	Pushare::Util::unlock() if ($opt{lock});
	return %name;
}

sub update() {
	Pushare::Util::lock();
	local $_;
	my %name = list();
	open OUT, "> $namelist" or die "$namelist: open failed. $!";
	my %cache = reverse(%name);
	while (glob ("$updir/*")) {
		my $file = $_;
		$file =~ s|$updir/||;
		$file = Jcode->new($file)->h2z->utf8;
		$file =~ s|[\\/]||g;
		unless ($cache{$file}) {
			my($IN, $OUT);
			my $pid = open2 $IN, $OUT, $Pushare::Config::command{md5sum}, $_;
			close $OUT;
			my $md5 = <$IN>;
			close $IN;
			wait;
			$md5 = substr $md5, 0, 32;
			system $Pushare::Config::command{cp}, $_, "$cachedir/$md5.dat";
			$name{$md5} = $file;
		}
	}
	my ($k, $v);
	while (($k, $v) = each %name) {
		if (-s "$cachedir/$k.dat") {
			print OUT "$k<>$v\n";
		}
	}
	close OUT;
	Pushare::Util::unlock();
}

sub delete($) {
	my($id) = @_;
	unlink "$cachedir/$id.dat" or return;
	Pushare::Util::lock();
	my %name = list();
	open OUT, "> $namelist" or die "$namelist: open failed. $!";
	my ($k, $v);
	while (($k, $v) = each %name) {
		if ($k ne $id) {
			print OUT "$k<>$v\n";
		}
	}
	close OUT;
	Pushare::Util::unlock();
}

sub add($$) {
	my($id, $name) = @_;
	$name =~ s|[\\/]||g;
	Pushare::Util::lock();
	open OUT, ">> $namelist" or die "$namelist: open failed. $!";
	print OUT "$id<>$name\n";
	close OUT;
	Pushare::Util::unlock();
}

sub name($) {
	my($id) = @_;
	my %name = list();
	return $name{$id};
}

sub convert($) {
	my($id) = @_;
	my $name = name($id);
	$name = "$id.dat" unless ($name);
	Jcode::convert(\$name, $Pushare::Config::jcode, "utf8");
	while (-e "$downdir/$name") {
		my @x = split /\./, $name;
		if ($x[-2] =~ /^\d+$/) {
			$x[-2]++;
		} else {
			my $tmp = $x[-1];
			$x[-1] = 1;
			push @x, $tmp;
		}
		$name = join ".", @x;
	}
	system $Pushare::Config::command{cp}, "$cachedir/$id.dat", "$downdir/$name";
}


1;
