#
# pushare - P2P file sharing system
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: $
#

package Pushare::NodeList;
use Pushare::Config;
use Pushare::Node;
use Pushare::Util;

#
# configration
#
$nodelist = $Pushare::Config::nodelist;

#
# show all nodes
#
sub list() {
	local $_;
	open IN, $nodelist or return ();
	my @node = ();
	while (<IN>) {
		chomp;
		my ($host, $port) = split /:/;
		my $n = new Pushare::Node($host, $port);
		push @node, $n;
	}
	return @node;
}

#
# show randmized all nodes
#
sub all() {
	local $_;
	my @node = list();
	for ($i=@node; $i>0; $i--) {
		my $t = int rand $i;
		($node[$i-1], $node[$t]) = ($node[$t], $node[$i-1]);
	}
	return @node;
}

#
# suggest random node
#
sub random() {
	local $_;
	my @node = list();
	if (@node) {
		my $node = $node[rand(0+@node)];
		return $node;
	} else {
		return undef;
	}
}

#
# who am i
#
sub myself() {
	my $port = $main::port;
	local $_;
	my @node = all();
	foreach (@node) {
		my @buf = $_->talk("/ping");
		if ((defined $buf[1])&&($buf[0] eq "PONG")) {
			return new Pushare::Node($buf[1], $port);
		}
	}
	return undef;
}

#
# add node to list
# args: node
#
sub add($) {
	my($node) = @_;
	my $str = $node->toString() . "\n";
	local $_;
	unless (include($node)) {
		Pushare::Util::lock();
		open OUT, ">> $nodelist" or die "$nodelist: open failed. $!";
		print OUT "$str";
		close OUT;
		Pushare::Util::unlock();
	}
}

#
# include node
# args: node
#
sub include($) {
	my($node) = @_;
	my $str = $node->toString();
	local $_;
	my @node = list();
	if (grep $_->toString() eq $str, @node) {
		return 1;
	} else {
		return undef;
	}
}

#
# count node
#
sub count() {
	my @node = list();
	return 0+@node;
}

#
# remove node from list
# args: node
#
sub remove($) {
	my($node) = @_;
	my $str = $node->toString();
	$str .= "\n";
	local $_;
	my @node;
	if (-e $nodelist) {
		open IN, $nodelist or exit;
		@node = <IN>;
		close IN;
	} else {
		return;
	}
	if (grep $_ eq "$str", @node) {
		Pushare::Util::lock();
		open OUT, "> $nodelist" or die "$nodelist: open failed. $!";
		foreach (@node) {
			print OUT if ($_ ne $str);
		}
		close OUT;
		Pushare::Util::unlock();
	}
}

#
# ping all nodes
#
sub pingall() {
	local $_;
	my @node = all();
	foreach (@node) {
		unless ($_->ping()) {
			remove($_);
		}
	}
}

#
# join network
# args: node, port
#
sub join($) {
	my ($node) = @_;
	my $port = $main::port;
	my $count = 0;
	while (1) {
		my @buf = $node->join();
		my $n = $node->toString();
		if ((@buf == 1)&&($buf[0] eq "WELCOME")) {
			add($node);
			last;
		} elsif ($count++ > $Pushare::Config::retrySearch) {
			last;
		} elsif ((@buf == 2)&&($buf[0] eq "WELCOME")) {
			add($node);
			my($h, $p) = split /:/, $buf[1], 2;
			$node = new Pushare::Node($n, $p);
		} else {
			last;
		}
	}
}

#
# connect initial node
# args: port
#
sub init() {
	my $port = $ENV{SERVER_PORT};
	my @initNode = @Pushare::Config::initNode;
	unshift @initNode, $main::initNode if ($main::initNode);
	local $_;
	my $inode;
	foreach (@initNode) {
		my($h, $p) = split /:/, $_, 2;
		$inode = new Pushare::Node($h, $p);
		if ($inode->ping()) {
			Pushare::NodeList::join($inode);
			last;
		}
	}
	my $myself = myself();
	remove($myself) if ($myself);
	my @node = all();
	my %done = ();
	return if (@node == 0);

	$done{$_->toString()} = 1 foreach (@node);
	while (1) {
		my $node = $node[rand(0+@node)];
		my @buf  = $node->talk("/node");
		my $str = "";
		if ((@buf > 0)&&($buf[0] =~ m|:\d+/[^: ]+$|)) {
			$str = $buf[0];
			my($h, $p) = split /:/, $str, 2;
			$node = new Pushare::Node($h, $p);
			push @node, $node unless ($done{$str});
		}
		if (($done{$str}++ > $Pushare::Config::retrySearch)||
				(@node >= $Pushare::Config::nodes)) {
			last;
		}
	}

	foreach (@node) {
		unless (include($_)) {
			my($h, $p) = split /:/, $_, 2;
			Pushare::NodeList::join($_);
		}
	}
	remove($myself) if ($myself);
	if (count() >= $Pushare::Config::nodes) {
		$inode->bye();
		remove($inode);
	}
	while (count() > $Pushare::Config::nodes) {
		my $node = random();
		$node->bye();
		remove($node);
	}
}

#
# relay distribution
#
sub relayDist($) {
	my($message) = @_;
	return if (rand() < $Pushare::Config::ignore);
	my @node = list();
	local $_;
	foreach (@node) {
		$_->talk("/dist?$message");
	}
}

1;
