#!/usr/bin/perl -w
#
# pushare - P2P file sharing system
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: $
#

use lib qw(.);
use Socket;
use Pushare::Cache;
use Pushare::Config;
use Pushare::Gateway;
use Pushare::Node;
use Pushare::NodeList;
use Pushare::Search;

$SIG{CHLD} = "IGNORE";

#
# default value
#
$port = $Pushare::Config::port;
$timeout = $Pushare::Config::timeout;
$cachedir = $Pushare::Config::cachedir;
$bufsize  = $Pushare::Config::bufsize;

while (@ARGV) {
	$_ = shift @ARGV;
	if ($_ eq "-p") {
		$port = shift @ARGV;
	} elsif ($_ eq "-i") {
		$initNode = shift @ARGV;
	}
}



#------------------------------------------------------------------#

sub dprint(@) {
	print STDOUT @_, "\n";
}

sub xdie(@) {
	if ((defined $pid_client)&&($pid_client>0)) {
		kill "TERM", $pid_client;
		dprint "terminate $pid_client";
	}
	die @_;
}

sub printHeader($) {
	my ($stat) = @_;
	my $ver = "0.1-alpha1";
	local $, = "\r\n";
	alarm $timeout;
	print "HTTP/1.0 $stat",
	      "Server: pushare/$ver",
	      "Connection: Close",
	      "";
	alarm 0;
}

sub print404() {
	printHeader("404 Not Found");
	print "Content-Length: 0\r\n",
	      "\r\n";
}

sub encode($) {			# from YukiWiki
	my ($s) = @_;
	my $encoded = '';
	foreach my $ch (split(//, $s)) {
		if ($ch =~ /[A-Za-z0-9.]/) {
			$encoded .= $ch;
		} else {
			$encoded .= '%' . sprintf("%02X", ord($ch));
		}
	}
	return $encoded;
}

sub decode($) {			# from YukiWiki
	my ($s) = @_;
	$s =~ tr/+/ /;
	$s =~ s/%([A-Fa-f0-9][A-Fa-f0-9])/pack("C", hex($1))/eg;
	return $s;
}

sub args($) {
	local $_ = $_[0];
	local @_ = split /&/;
	my %arg = ();
	foreach (@_) {
		local @_ = split /=/;
		if (defined $_[1]) {
			$_[1] =~ s/\+/ /g;
			$_[1] =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
			$arg{$_[0]} = $_[1];
		} else {
			$arg{$_[0]} = "";
		}
	}
	return %arg;
}

sub dist_include($) {
	my($id) = @_;
	local $_;
	if (open IN, $Pushare::Config::distlist) {
		while (<IN>) {
			if (/$id/) {
				close IN;
				return 1;
			}
		}
		close IN;
	}
	return 0;
}

#------------------------------------------------------------------#
#
# httpd
#
sub httpd_main() {
	$SIG{TERM} = $SIG{HUP} = $SIG{CHLD} = "DEFAULT";
	$SIG{ALRM} = sub { die "$remote_addr: server timeout" };

	# parse HTTP request
	my($str_request, $http_referer, $http_user_agent) = ("", "", "");
	alarm $timeout;
	while (<CLIENT>) {
		s/[\n|\r]//g;
		if ($_ eq "") {
			last;
		} elsif (/^(\S+)\s+(\S+)/ && (!defined $request_method)) {
			$str_request    = $_;
			$request_method = $1;
			$request_uri	= $2;
			$request_uri	=~ s|^//+|/|g;

		} elsif (/^Host:\s+/i) {
			$http_host = $';

		} elsif (/^Range:\s*bytes\s*=\s*(\d+)\s*-/) {
			$range = $1;

		} elsif (/^Referer:\s+/i) {
			$http_referer = $';

		} elsif (/^User-Agent:\s+/i) {
			$http_user_agent = $';

		}
	}
	alarm 0;

	if ($request_uri =~ m|^[a-z]+://([^/]+)|) {
		$http_host   = $1;
		$request_uri = $';
	}
	if ($request_uri eq "") {
		$request_uri = "/";
	}
	if ($request_uri eq "/") {
		$path = "/";
		$query_string = "";
	} elsif ($request_uri =~ /\?/) {
		$path = $`;
		$query_string = $';
	} else {
		$path = $request_uri;
		$query_string = "";
	}

	# output logs
	my $date = localtime;
	dprint "$remote_addr<>$date<>$str_request<>$http_referer<>$http_user_agent";

	if ($request_method eq "OPTIONS") {
		printHeader("200 OK");
		print "Allow: GET, HEAD, OPTIONS\n",
		      "Content-Length: 0\r\n\r\n";

	} elsif ($request_method eq "HEAD") {
		printHeader("200 OK");
		print "Content-Length: 0\r\n",
		      "\r\n";

	} elsif ($request_method !~ /^(GET|HEAD)$/) {
		printHeader("501 Not Implemented");
		print "\r\n";

	} elsif ($path eq "/ping") {
		printHeader("200 OK");
		print "Content-Type: text/plain\r\n",
		      "\r\n",
		      "PONG\n",
		      "$remote_addr\n";

	} elsif ($path eq "/node") {
		my $node = Pushare::NodeList::random();
		if ($node) {
			printHeader("200 OK");
			print "Content-Type: text/plain\r\n",
			      "\r\n",
			      $node = $node->toString(), "\n";
		} else {
			print404();
			return;
		}

	} elsif ($path =~ m|^/get/([0-9a-f]{32})$|) {
		my $file = $1;
		unless ((-f "$cachedir/$file.dat")
			&& (dist_include($file))
			&& (open IN, "$cachedir/$file.dat")) {
			print404();
			return;
		}
		my $size = (-s "$cachedir/$file.dat");
		if (defined $range) {
			my $all = $size;
			$size -= $range;
			printHeader("206 Partial content");
			print "Content-Range: bytes $range-" . ($all-1)  . "/$all\r\n";
		} else {
			printHeader("200 OK");
		}
		print "Content-Length: $size\r\n",
		      "Content-Type: application/octet-stream\r\n",
		      "\r\n";
		my $buf;
		alarm $timeout;
		while (read IN, $buf, $bufsize) {
			print $buf;
			alarm $timeout;
		}

	} elsif ($path eq "/join") {
		my %arg = args($query_string);
		unless ($arg{port}) {
			print404();
			return;
		}
		my $node = new Pushare::Node($remote_addr, $arg{port});
		unless ($node->ping()) {
			print404();
			return;
		} elsif (Pushare::NodeList::include($node) ||
			(Pushare::NodeList::count() < $Pushare::Config::nodes)) {
			Pushare::NodeList::add($node);
			printHeader("200 OK");
			print "Content-Type: text/plain\r\n",
			      "\r\n",
			      "WELCOME\n";
		} else {
			my $suggest = Pushare::NodeList::random();
			Pushare::NodeList::remove($suggest);
			$suggest->bye();
			printHeader("200 OK");
			print "Content-Type: text/plain\r\n",
			      "\r\n",
			      "WELCOME\n",
			      $suggest->toString(), "\n";
			Pushare::NodeList::add($node);
		}

	} elsif ($path eq "/bye") {
		my %arg = args($query_string);
		unless ($arg{port}) {
			print404();
			return;
		}
		my $node = new Pushare::Node::($remote_addr, $arg{port});
		Pushare::NodeList::remove($node);
		printHeader("200 OK");
		print "Content-Type: text/plain\r\n",
		      "\r\n",
		      "BYEBYE\n";

	} elsif ($path eq "/dist") {
		local $_;
		my %arg = args($query_string);
		unless ($arg{id} && $arg{host} && $arg{port}) {
			print404();
			return;
		}
		my $date = time;

		my @ignore = ();
		if (open IN, $Pushare::Config::ignorelist) {
			@ignore = <IN>;
			close IN;
		}
		if (dist_include($arg{id})) {

		} elsif (-s "$cachedir/$arg{id}.dat") {
			Pushare::Util::addDist($date, $arg{id});
			my $myself = Pushare::NodeList::myself();
			$myself = $$myself{host};
			$query_string =~ s/host=$arg{host}/host=$myself/;
			$query_string =~ s/port=$arg{port}/port=$port/;
			Pushare::NodeList::relayDist($query_string);

		} elsif (grep $_ eq $arg{id}, @ignore) {
			Pushare::Util::addDist($date, $arg{id});
			Pushare::NodeList::relayDist($query_string);

		} elsif (Pushare::Search::check(%arg)) {
			my $id = $arg{id};
			Pushare::Util::addDist($date, $id);
			my $node = new Pushare::Node($remote_addr, $arg{port});
			if ($node->talk("/get/$id", file=>$id)) {
				my $myself = Pushare::NodeList::myself();
				$myself = $$myself{host};
				my $name = decode($arg{name});
				$query_string =~ s/host=$arg{host}/host=$myself/;
				$query_string =~ s/port=$arg{port}/port=$port/;
				Pushare::Util::addIgnore($id);
				Pushare::Cache::add($id, $name);
				Pushare::Cache::convert($id);
				Pushare::NodeList::relayDist($query_string);
			} else {
				Pushare::Util::delDist($date, $id);
			}

		} else {
			Pushare::Util::addDist($date, $arg{id});
			Pushare::NodeList::relayDist($query_string);
		}
		printHeader("200 OK");
		print "Content-Type: text/plain\r\n",
		      "\r\n",
		      "OK\n";

	} elsif (($path eq "/") || ($path =~ m|^/gw|)) {
		Pushare::Gateway::main();

	} else {
		printHeader("404 Not Found");
		print "\r\nBad arguments or No data.\n",
		      "Try later.\n";
	}
}

sub httpd() {
	# initialize daemon
	socket CLIENT_WAITING, PF_INET, SOCK_STREAM, 0 or xdie "failed to open socket. $!\n";
	setsockopt CLIENT_WAITING, SOL_SOCKET, SO_REUSEADDR, 1 or xdie "failed to setsockopt. $!\n";
	bind CLIENT_WAITING, pack_sockaddr_in($port, INADDR_ANY) or xdie "failed to bind. $!\n";
	listen CLIENT_WAITING, SOMAXCONN or xdie "failed to listen: $!\n";

	# run daemon
	while (1) {
		sleep 1 if (rand() < 1/$Pushare::Config::children);
		unless ($paddr = accept(CLIENT, CLIENT_WAITING)) {
			sleep 1;
			next;
		}
		my ($client_port, $client_iaddr) = unpack_sockaddr_in $paddr;
		local $remote_addr = inet_ntoa $client_iaddr;
		select CLIENT; $|=1;

		if (! defined($pid = fork)) {
			dprint "server process: fork failed. $!";
			close CLIENT;
			next;

		} elsif ($pid) {
			close CLIENT;
			next;

		} else {
			httpd_main();
			shutdown CLIENT, 2;
			close CLIENT;
			exit;
		}
	}
}

#------------------------------------------------------------------#
#
# client process
#
sub client_save(%) {
	my(%time) = @_;
	local $, = "\n";
	Pushare::Util::lock();
	open OUT, "> $Pushare::Config::client"
		or die "Failed to open $Pushare::Config::client. $!";
	print OUT $time{ping}, $time{join}, $time{init}, $time{dist}, "";
	close OUT;
	Pushare::Util::unlock();
}

sub client_main() {
	$SIG{ALRM} = sub { die "client timeout" };
	alarm $Pushare::Config::timeout_client * 60;
	my %time = ();
	Pushare::Cache::update();
	if (open IN, $Pushare::Config::client) {
		$time{ping} = <IN>;
		$time{join} = <IN>;
		$time{init} = <IN>;
		$time{dist} = <IN>;
		chomp($time{ping}, $time{join}, $time{init}, $time{dist});
		close IN;
	} else {
		($time{ping}, $time{join}, $time{init}, $time{dist}) = (0, 0, 0, 0);
	}
	if ((time - $time{ping}) >= $Pushare::Config::pingfrec * 60) {
		$time{ping} = time;
		client_save(%time);
		Pushare::NodeList::pingall();
		dprint "ping";
	}
	my $nc = Pushare::NodeList::count();
	if ((time - $time{init}) >= $Pushare::Config::initfrec * $nc * 60) {
		$time{init} = time;
		client_save(%time);
		Pushare::NodeList::init();
		dprint "init";
	}
	if ((time - $time{join}) >= $Pushare::Config::joinfrec * 60) {
		client_save(%time);
		my @node = Pushare::NodeList::all();
		foreach (@node) {
			my $n = $_->toString();
			Pushare::NodeList::join($_);
		}
		$time{join} = time;
		dprint "join";
	}
	if ((time - $time{dist}) >= $Pushare::Config::distfrec * 60) {
		client_save(%time);
		my @file = glob("$cachedir/*.dat");
		return unless (@file);
		my $i;
		for ($i=$#file; $i>=0; $i--) {
			my $t = int rand $i;
			($file[$i], $file[$t]) = ($file[$t], $file[$i]);
		}
		my $file;
		foreach $file (@file) {
			my $id	 = $file;  $id =~ s|$cachedir/||;  $id =~ s|\.dat$||;
			next if (dist_include($id));
			my $size = (-s $file);
			my $name = Pushare::Cache::name($id);
			if ($name) {
				$name = encode($name);
			} else {
				$name = "";
			}
			my $myself = Pushare::NodeList::myself();
			$myself = $$myself{host};
			Pushare::Util::addDist(time, $id);
			Pushare::NodeList::relayDist("host=$myself&port=$port&id=$id&name=$name");
			last;
		}
	}
}

sub client() {
	$SIG{TERM} = $SIG{HUP} = "DEFAULT";
	$SIG{CHLD} = "IGNORE";
	mkdir $Pushare::Config::updir, 0777 unless (-d $Pushare::Config::updir);
	mkdir $Pushare::Config::downdir, 0777 unless (-d $Pushare::Config::downdir);
	mkdir $Pushare::Config::cachedir, 0777 unless (-d $Pushare::Config::cachedir);
	unlink $Pushare::Config::client;
	sleep 5;

	while (1) {
		my $pid;
		if (! defined ($pid = fork)) {
			dprint "client process: fork failed.";
			sleep $Pushare::Config::clientfrec * 60;
			next;

		} elsif ($pid) {
			sleep 10;
			my $t = $Pushare::Config::clientfrec * 60
			      + int($Pushare::Config::timeError * Pushare::Util::normal());
			$t = $Pushare::Config::clientfrec * 60 if ($t <= 0);
			sleep $t;
			next;

		} else {
			client_main();
			exit;
		}
	}
}


#------------------------------------------------------------------#

if (($pid_client=fork) == 0) {
	client();
}
httpd();
